@@ -1,13 +1,14 @@
 package Centrifugo::Client;
 
-our $VERSION = "1.03";
+our $VERSION = "1.04";
 
 use Exporter;
 our @ISA = qw(Exporter);
-our @EXPORT = qw();
+our @EXPORT = qw(generate_token);
 
 use Carp qw( croak );
-use AnyEvent::WebSocket::Client 0.12;
+use AnyEvent::WebSocket::Client 0.40; # Version needed for reason when close. See https://github.com/plicease/AnyEvent-WebSocket-Client/issues/30
+use AnyEvent::HTTP;
 use JSON;
 
 =head1 NAME
@@ -58,8 +59,11 @@ use warnings;
 or
 
 	my $client = Centrifugo::Client->new( $URL,
-	   debug => 'true',           # if true, some informations are written on STDERR
-	   ws_params => {             # These parameters are passed to AnyEvent::WebSocket::Client->new(...)
+	   debug => 'true',          # If true, some informations are written on STDERR
+	   authEndpoint => "...",    # The full URL used to ask for a key to subscribe to private channels
+	   max_alive_period => 30,   # If max_alive_period has passed since last communication with server, a PING is send (default 0)
+	   refresh_period => 5,      # Check frequency for max_alive_period (default 10s)
+	   ws_params => {            # These parameters are passed to AnyEvent::WebSocket::Client->new(...)
 			 ssl_no_verify => 'true',
 			 timeout => 600
 		  },
@@ -73,7 +77,10 @@ sub new {
 	bless($this, $class);
 	$this->{WS_URL} = $ws_url;
 	$this->{DEBUG} = $params{debug} && uc($params{debug})ne'FALSE';
+	$this->{AUTH_URL} = $params{authEndpoint} || "/centrifuge/auth/";
 	$this->{WEBSOCKET} = AnyEvent::WebSocket::Client -> new( %{$params{ws_params}} );
+	$this->{MAX_ALIVE} = $params{max_alive_period} || 0;
+	$this->{REFRESH} = $params{refresh_period} || 10;
 	return $this;
 }
 
@@ -120,17 +127,18 @@ sub connect {
 			params => \%PARAMS
 		};
 		
-		print STDERR "Centrifugo::Client : WS > $CONNECT\n" if $this->{DEBUG};
+		print STDERR "Centrifugo::Client : WebSocket > $CONNECT\n" if $this->{DEBUG};
 		$this->{WSHANDLE}->on(each_message => sub {
 			my($loop, $message) = @_;
-			print STDERR "Centrifugo::Client : WS < $message->{body}\n" if $this->{DEBUG};
-			my $body = decode_json($message->{body});
+			print STDERR "Centrifugo::Client : R< WS : $message->{body}\n" if $this->{DEBUG};
+			$this->{last_alive_message} = time();
+			my $fullbody = decode_json($message->{body});
 			# Handle a body containing {response}
-			if (ref($body) eq 'HASH') {
-				$body = [ $body ];
+			if (ref($fullbody) eq 'HASH') {
+				$fullbody = [ $fullbody ];
 			}
 			# Handle a body containing [{response},{response}...]
-			foreach my $info (@$body) {
+			foreach my $info (@$fullbody) {
 				my $uid = $info->{uid};
 				my $method = $info->{method};
 				my $error = $info->{error};
@@ -141,6 +149,20 @@ sub connect {
 						$this->{CLIENT_ID} = $body->{client};
 						print STDERR "Centrifugo::Client : CLIENT_ID=".$this->{CLIENT_ID}."\n" if $this->{DEBUG};
 					}
+					if ($this->{MAX_ALIVE}) {
+						# Creates the timer to send periodic ping
+						$this->{alive_handler} = AnyEvent->timer(
+							after => $this->{REFRESH},
+							interval => $this->{REFRESH},
+							cb => sub {
+								my $late = time() - $this->{last_alive_message};
+								if ($late > $this->{MAX_ALIVE}) {
+									print STDERR "Sending ping (${late}s without message)\n" if $this->{DEBUG};
+									$this->ping();
+								}
+							}
+						);
+					}
 				}
 				# Call the callback of the method
 				my $sub = $this->{ON}->{$method};
@@ -154,20 +176,20 @@ sub connect {
 			}
 		});
 
-		unless ($^O=~/Win/i) {
-			# This event seems to be unrecognized on Windows (?)
-			$this->{WSHANDLE}->on(parse_error => sub {
-				my($loop, $error) = @_;
-				warn "Error in Centrifugo::Client : $error";
-				$this->{ON}->{'error'}->($error) if $this->{ON}->{'error'};
-			});
-		}
+		$this->{WSHANDLE}->on(parse_error => sub {
+			my($cnx, $error) = @_;
+			warn "Error in Centrifugo::Client : $error";
+			$this->{ON}->{'error'}->($error) if $this->{ON}->{'error'};
+		});
 
 		# handle a closed connection...
 		$this->{WSHANDLE}->on(finish => sub {
-			my($loop) = @_;
-			print STDERR "Centrifugo::Client : Connection closed\n" if $this->{DEBUG};
-			$this->{ON}->{'ws_closed'}->() if $this->{ON}->{'ws_closed'};
+			my($cnx) = @_;
+			my $reason = $cnx->close_reason();
+			
+			print STDERR "Centrifugo::Client : Connection closed (reason=$reason)\n" if $this->{DEBUG};
+			$this->{ON}->{'ws_closed'}->($reason) if $this->{ON}->{'ws_closed'};
+			undef $this->{alive_handler};
 			undef $this->{WSHANDLE};
 			undef $this->{CLIENT_ID};
 		});
@@ -208,8 +230,7 @@ sub publish {
 		method => 'publish',
 		params => \%PARAMS
 	};
-	print STDERR "Centrifugo::Client : WS > $PUBLISH\n" if $this->{DEBUG};
-	$this->{WSHANDLE}->send( $PUBLISH );
+	$this->_send_message($PUBLISH);
 	return $uid;
 }
 
@@ -228,7 +249,9 @@ sub disconnect {
 
 =head1 FUNCTION subscribe - allows to subscribe on channel after client successfully connected.
 
-$client->subscribe( channel => $channel, [ uid => $uid ] );
+$client->subscribe( channel => $channel, [ client => $clientId  ,] [ uid => $uid ,] );
+
+If the channel is private (starts with a '$'), then a request to $this->{AUTH_URL} is done automatically to get the channel key. In that case, the 'client' parameter is mandatory.
 
 This function returns the UID used to send the command to the server. (a random string if none is provided)
 
@@ -236,21 +259,43 @@ This function returns the UID used to send the command to the server. (a random
 
 sub subscribe {
 	my ($this, %PARAMS) = @_;
-	return _channel_command($this,'subscribe',%PARAMS);
+	my $channel = $PARAMS{channel};
+	return _channel_command($this,'subscribe',%PARAMS) unless $channel=~/^\$/;
+	# If the channel is private, then an API-call to /centrifuge/auth/ must be done
+	croak "'client' parameter is mandatory to subscribe to private channels in Centrifugo::Client->subscribe(...)" unless $PARAMS{client};
+
+	# Request a channel key
+	my $data = encode_json {
+		client => $PARAMS{client},
+		channels => [ $channel ]
+	};
+	my $URL = $this->{AUTH_URL};
+	http_post $URL, $data, 
+		headers => {
+			contentType => "application/json"
+		},
+		sub {
+		  my ($data, $headers) = @_;
+		  warn "Couldn't connect to $URL : Status=".$headers->{Status} and return unless $headers->{Status}==200;
+		  my $result = decode_json $data;
+		  my $key = $result->{$channel}->{sign};
+		  $PARAMS{sign} = $key;
+		  # The request is now complete : {channel:"...", client:"...", sign:"..."}
+		  return _channel_command($this,'subscribe',%PARAMS);
+	   };
 }
 
 sub _channel_command {
 	my ($this,$command,%PARAMS) = @_;
-	my $channel = $PARAMS{'channel'};
-	croak("Missing channel in Centrifugo::Client->$command(...)") unless $channel;
+#	my $channel = $PARAMS{'channel'};
+#	croak("Missing channel in Centrifugo::Client->$command(...)") unless $PARAMS{'channel'};
 	my $uid = $PARAMS{'uid'} || _generate_random_id();
 	my $MSG = encode_json {
 		UID => $uid ,
 		method => $command,
-		params => { channel => $channel }
+		params => \%PARAMS
 	};
-	print STDERR "Centrifugo::Client : WS > $MSG\n" if $this->{DEBUG};
-	$this->{WSHANDLE}->send($MSG);
+	$this->_send_message($MSG);
 	return $uid;
 }
 
@@ -267,7 +312,7 @@ sub unsubscribe {
 	return _channel_command($this,'unsubscribe',%PARAMS);
 }
 
-=head1 FUNCTION presence – allows to ask server for channel presence information.
+=head1 FUNCTION presence - allows to ask server for channel presence information.
 
 $client->presence( channel => $channel, [ uid => $uid ] );
 
@@ -280,7 +325,7 @@ sub presence {
 	return _channel_command($this,'presence',%PARAMS);
 }
 
-=head1 FUNCTION history – allows to ask server for channel presence information.
+=head1 FUNCTION history - allows to ask server for channel presence information.
 
 $client->history( channel => $channel, [ uid => $uid ] );
 
@@ -293,7 +338,7 @@ sub history {
 	return _channel_command($this,'history',%PARAMS);
 }
 
-=head1 FUNCTION ping – allows to send ping command to server, server will answer this command with ping response.
+=head1 FUNCTION ping - allows to send ping command to server, server will answer this command with ping response.
 
 $client->ping( [ uid => $uid ] );
 
@@ -308,14 +353,11 @@ sub ping {
 		UID => $uid ,
 		method => 'ping'
 	};
-	print STDERR "Centrifugo::Client : WS > $MSG\n" if $this->{DEBUG};
-	$this->{WSHANDLE}->send($MSG);
+	$this->_send_message($MSG);
 	return $uid;
 }
 
-=head1 FUNCTION on
-
-Register a callback for the given event.
+=head1 FUNCTION on - Register a callback for the given event.
 
 Known events are 'message', 'connect', 'disconnect', 'subscribe', 'unsubscribe', 'publish', 'presence', 'history', 'join', 'leave',
 'refresh', 'ping', 'ws_closed', 'ws_error'
@@ -337,9 +379,9 @@ sub on {
 	$this;
 }
 
-=head1 FUNCTION client_id
+=head1 FUNCTION client_id - return the client_id if it is connected to Centrifugo and the server returned this ID (which is not the case on the demo server).
 
-$client->client_id() return the client_id if it is connected to Centrifugo and the server returned this ID (which is not the case on the demo server).
+$client->client_id() 
 
 =cut
 
@@ -349,12 +391,54 @@ sub client_id {
 }
 
 
+=head1 FUNCTION generate_token - return the private token that must be used to connect a client to Centrifugo.
+
+$key = Centrifugo::Client::generate_token($secret, $user, $timestamp [,$info])
+
+INPUT : $secret is the private secret key, only known by the server.
+
+        $user is the user name.
+        
+        $timestamp is the current timestamp.
+        
+        $info is a JSON encoded string.
+
+The same function may be used to generate a private channel key :
+
+    $key = generate_token($secret, $client, $channel [,$info])
+
+INPUT : $client is the client_id given when connected to Centrifugo.
+
+        $channel is the name of the channel (should start with a '$' as it is private).
+
+And to sign each request to access to the HTTP API :
+
+    $sign = generate_token($self, $data)
+
+INPUT : $data is a JSON string with your API commands
+
+=cut
+
+sub generate_token {
+	my ($secret, @infos)=@_;
+	my $info = join'', @infos;
+	use Digest::SHA qw( hmac_sha256_hex );
+	return hmac_sha256_hex( $info, $secret );
+}
+
 ##### (kinda)-private functions
 
+sub _send_message {
+	my ($this,$MSG)=@_;
+	print STDERR "Centrifugo::Client : S> WebSocket : $MSG\n" if $this->{DEBUG};
+	$this->{WSHANDLE}->send($MSG);
+	
+}
+
+
 # Generates a random Id for commands
 sub _generate_random_id {
 	my @c = ('a'..'z','A'..'Z',0..9);
 	return join '', @c[ map{ rand @c } 1 .. 12 ];
 }
-
 1;
